/*
  Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
*/

#ifndef GEOMODELKERNEL_GEOPOLYHEDRONGENERIC_H
#define GEOMODELKERNEL_GEOPOLYHEDRONGENERIC_H

/**
 * @class: GeoPolyhedronGeneric
 *
 * @brief This shape represents a generic polyhedron, described by an array of vertices
 *  and faces defined by a set of ordered vertices
 */

#include <vector>
#include "GeoModelKernel/GeoShape.h"

class GeoPolyhedronGeneric : public GeoShape {
 public:
  // Default constructor 
  GeoPolyhedronGeneric(){};

  // Constructor 
  GeoPolyhedronGeneric(const GeoShape*, const GeoTrf::Transform3D& trf = GeoTrf::Transform3D::Identity(), int polyhedrize = 0);

  // Copy constructor
  GeoPolyhedronGeneric(const GeoPolyhedronGeneric& gpg); 

  //    Returns the volume of the shape, for mass inventory
  virtual double volume() const;

  //    Returns the bonding box of the shape
  virtual void extent (double& xmin, double& ymin, double& zmin,
                       double& xmax, double& ymax, double& zmax) const;

  //    Returns true if the shape contains the point, false otherwise.
  virtual bool contains (double x, double y, double z) const;

  //    Returns true if the shape contains the point, false otherwise.
  virtual bool contains (GeoTrf::Vector3D point, double precision = 1.e-6, bool debugMode = false) const;

  //    Returns the shape type, as a string
  virtual const std::string& type() const {
    return getClassType();
  }

  //    Returns the shape type, as a coded integer
  virtual ShapeType typeID() const{
    return getClassTypeID();
  }

  //    For type identification
  static const std::string& getClassType() {
    return s_classType;
  }

  //    For type identification
  static ShapeType getClassTypeID() {
    return s_classTypeID;
  }

  //    Executes a GeoShapeAction
  virtual void exec(GeoShapeAction *action) const;

  // Define vertices 
  void addVertices(std::vector<GeoTrf::Vector3D> vertices);

  //    Add another facet to the polyhedron. A minimum of four
  //    faces are required to create a valid solid.
  void addFace(std::vector<size_t> face);

  // normals can be uploaded or calculated 
  void addNormal(GeoTrf::Vector3D n);

  // finalize GPG 
  bool construct(bool debugMode = false);

  //    Returns the number of vertices of the polyhedron
  unsigned int getNVertices() const { return m_Vertices.size(); }

  //    Returns the number of (boundary) faces of the polyhedron
  unsigned int getNFaces() const { return m_Faces.size(); }

  //  check of the hermeticity of the polyhedron (all edge need to be mirrored), plus at least four faces required, false otherwise
  bool isValid ( bool debugMode = false) const;

  void setStatus( int status) { m_status = status;}
  int getStatus() const { return m_status; }  

  //    Returns coordinates of the specified vertex
  GeoTrf::Vector3D getVertex(unsigned int i) const {
    return m_Vertices.at(i);
  }

  //    Returns definition of the specified face
  std::vector<size_t> getFace(unsigned int i) const {
    return m_Faces.at(i);
  }

  //  return normal ( pointing outwards )
  GeoTrf::Vector3D faceNormal(unsigned int i) const;

  //  true for point laying on the face
  bool insideFace(unsigned int i, GeoTrf::Vector3D point, double precision = 1.e-6, bool debugMode = false) const;

  // method to intersect ( extra-polyhedron) edge with face
  std::vector< std::pair< double, std::pair<size_t,size_t> > > edgeFaceIntersect( int i, GeoTrf::Vector3D eInit, GeoTrf::Vector3D edge ) const;

  GeoTrf::Vector3D getRefPoint(int face = -1) const;

  int getOrigin() const;
  void setOrigin( int );

  void cleanAfterBooleanOperation();

  //    Returns true 
  virtual bool isPolyhedron () const {
    return true;
  }
  
  virtual ~GeoPolyhedronGeneric() = default;
protected:

 private:
  static const std::string s_classType;
  static const ShapeType s_classTypeID;
  GeoTrf::Transform3D m_transform{GeoTrf::Transform3D::Identity()};

  std::vector<GeoTrf::Vector3D> m_Vertices{};
  std::vector< std::vector<size_t> > m_Faces{};
  GeoTrf::Vector3D m_cog{GeoTrf::Vector3D(0.,0.,0.)};
  double m_span{0.};
  std::vector<GeoTrf::Vector3D> m_face_cog;
  std::vector<double> m_face_span;
  mutable std::vector<GeoTrf::Vector3D> m_Normals{};

  bool intersectLines(GeoTrf::Vector3D A, GeoTrf::Vector3D a, GeoTrf::Vector3D B, GeoTrf::Vector3D b, double& da, double& db, double& dx) const;

  // fill reference span for volume and face : run automatically for transcript from GeoShape
  void fillSpans();

  // retrieve cached span value
  double getRefSpan(int face = -1) const;

  // for debugging purposes
  int m_origin{-1};

  // status : -1 unchecked, 0 valid (may be empty), 1 non-polyhedrizable component, 2 failure to construct
  int m_status{-1};
  
};

#endif
