/*
  Copyright (C) 2002-2024 CERN for the benefit of the ATLAS collaboration
*/

#ifndef GEOMODELKERNEL_GEOPOLYHEDRONMANIPULATOR_H
#define GEOMODELKERNEL_GEOPOLYHEDRONMANIPULATOR_H

/**
 * @class GeoPolyhedronManipulator
 *
 * @brief This helper class collects methods for analytical and numerical processing of boolean shapes involving polyhedra.
 *
 */

#include <GeoModelKernel/GeoDefinitions.h>
#include "GeoModelKernel/GeoShape.h"
#include "GeoModelKernel/GeoShapeShift.h"
#include "GeoModelKernel/GeoPolyhedronGeneric.h"
#include <string>

class GeoShapeIntersection;
class GeoShapeUnion;
class GeoShapeSubtraction;

struct ShiftedGeoShape {
  const GeoShape*        shape{nullptr};
  const GeoPolyhedronGeneric* phed{nullptr};
  GeoTrf::Transform3D    transform;
  double                 volume{-1};   // -1  not calculated/not analytical
  int                    polyhedrize{16};   // >0 polyhedrize
  // constructor 
  ShiftedGeoShape(const GeoShape* inputShape) {
    shape = inputShape;
    transform = GeoTrf::Transform3D::Identity();
    // resolve (nested) shift transform
    while ( shape->type()=="Shift" ) {
      const GeoShapeShift* shift = dynamic_cast<const GeoShapeShift*> (shape);  
      transform = shift->getX()*transform;
      shape = shift->getOp();
    } 
    if ( !(shape->type()=="Union" || shape->type()=="Subtraction" || shape->type()=="Intersection") ) {
      volume = shape->volume();
      phed = new GeoPolyhedronGeneric(shape,GeoTrf::Transform3D::Identity(),polyhedrize);
      if (!phed->isValid()) phed=nullptr;     // temporary / to deal with non-polyhedrized shapes
    }
  }
};


struct NPL_vertex{
  int index;  bool shared{false}; int overlapping_face{-1}; bool used{false}; int onEdge{-1}; bool onEdgeDummy{false};
  int shared_adjacent{-1}; bool oriented{false}; // int alternative_face;
  bool onEdgeProtected{false};  // vertex with protected line connection
  std::vector<int> active_lines;
  std::pair<int,int>  faces;  // oriented : (incoming adjacent, outgoing hit face : edge->edge, edge->line, line->edge)
  std::pair<int,int>  faces_shared_loop; // non-oriented : line->line
  NPL_vertex() { overlapping_face = -1; used = false; shared_adjacent = 0; onEdgeDummy = false; oriented = false;}
};

struct NPL_edge{
  std::pair<int,int>  endpoints{std::make_pair<int>(-1,-1)};
  std::pair<int,int>  endpoint_faces{std::make_pair<int>(-1,-1)};
  std::pair<int,int>  endpoint_hitvx{std::make_pair<int>(-1,-1)};
  int adjacent_face{-1}; int overlap{-1}; int generating_edge{-1};
  bool used{false}; bool shared{false}; int symmetrized{-1};
  bool protectedLine{false}; int dummy_overlap{-1};
  // defoult constructor
  NPL_edge(){} 
  // constructor 
  NPL_edge(std::pair<int,int> ep, int adjacent){ endpoints = ep; adjacent_face = adjacent; }
};

struct NPL_ordered{
  std::vector<size_t> vtx;
  std::vector<int> edges;
  int top_overlap{-1}; bool shared{false}; bool oriented{false}; bool guess_orientation{false};
  void reorder(size_t a) {
    while ( vtx.front()!= a) {
      vtx.push_back(vtx.front());
      vtx.erase(vtx.begin());
      edges.push_back(edges.front());
      edges.erase(edges.begin());
    }
  }
  void reverse(std::vector<NPL_edge> plane_edges) {
    std::vector<size_t> vtx_tmp;
    std::vector<int> edges_tmp;
    for (auto iv : vtx) vtx_tmp.insert(vtx_tmp.begin(),iv);
    for (auto ie : edges) {
      edges_tmp.insert(edges_tmp.begin(),plane_edges[ie].symmetrized);
    }
    edges_tmp.push_back(edges_tmp.front()); edges_tmp.erase(edges_tmp.begin());    // reorder to match vtx 
    vtx = vtx_tmp;
    edges = edges_tmp;
  }
};

struct NPH_plane{
  int index_ph;
  std::vector<size_t> vertices;
  std::vector<size_t> edges;
  std::vector<NPL_vertex> hit_vertices;
  std::vector<std::pair< int, std::vector< NPL_vertex > > > active_planes; 
  std::vector<NPL_vertex> vlog;
  std::vector<NPL_vertex> vlog_ori;
  //std::vector<NPL_ordered> shared_subface;
  std::vector<NPL_ordered> subfaces;
  std::vector<NPL_edge> plane_edges;
  std::vector<NPL_edge> edge_loop_origin;
  GeoTrf::Vector3D normal;
  std::vector<int> overlapping;
  NPH_plane(int index) { index_ph = index;}
};

struct NPH_vertex{
  int index{-1};
  int index_ph{-1};
  bool shared{false};
  //int onDoublEdge{-1}; //int hitDoublEdge{-1};
  std::pair<int,int> hit_edge_edge;
  std::vector<std::pair<int,int> > edges;
  NPH_vertex() {hit_edge_edge = std::pair<int,int>(-1,-1);}
};

struct NPH_vtx{    // vertex created by intersection of edge with face or edge or vertex of another polyhedron
  int index; 
  int edge;  // -1 for original vertices, generating edge otherwise
  double distance;  // along generating edge
  int hitPlane; // index of intersected plane
  int hitEdge; // index of intersected edge
  int hitVertex; // index of intersected vertex
  std::pair<int,int> faces;
  NPH_vtx(int vtx_index) { index = vtx_index; hitPlane = -1; hitEdge = -1; hitVertex = -1; }  
};

struct NPH_edge{
  int index_ph; int edge_inverse;
  std::pair<int,int>  endpoints;
  std::pair<int,int>  endpoint_faces;
  std::pair<int,int>  faces;
  std::vector<int> inPlane;
  std::vector<int> doubleEdge;
  std::vector<NPH_vtx>  evtx;   // vector of vertices along edge, index+relative distance ( within (0,1) interval )
  // defoult constructor
  NPH_edge() { endpoints = std::make_pair<int>(-1,-1); faces = std::make_pair<int>(-1,-1); endpoint_faces = std::make_pair<int>(-1,-1); edge_inverse = -1; } 
  // constructor 
  NPH_edge(std::pair<int,int> ep){ endpoints = ep; faces = std::make_pair<int>(-1,-1); endpoint_faces = std::make_pair<int>(-1,-1); edge_inverse = -1; } 
};

struct NPH_line{
  std::pair<int,int>  faces;
  std::pair<int,int>  endpoints;
  std::vector<int>  evtx;   // vector of vertices along edge, index+relative distance ( within (0,1) interval )
  // constructor
  NPH_line(std::pair<int,int> ef){ faces = ef; } 
};

class GeoPolyhedronManipulator
{
 public:

  // steering routine for volume calculation with pre-defined precision
  double volumeBoolean( const GeoShape* shape, double tolerance, int mode) const;

  // resolve into generic polyhedron shape if possible, or deploy polyhedrization 
  GeoPolyhedronGeneric resolve( const GeoShape*, GeoTrf::Transform3D& trf, int polyhedrize = 0, int level = 0, bool debugMode = false );
  
  GeoPolyhedronGeneric* intersection( const GeoPolyhedronGeneric* phA, const GeoPolyhedronGeneric* phB, bool debugMode);
  GeoPolyhedronGeneric* subtraction( const GeoPolyhedronGeneric* phA, const GeoPolyhedronGeneric* phB, bool debugMode);
  GeoPolyhedronGeneric* uni( const GeoPolyhedronGeneric* phA, const GeoPolyhedronGeneric* phB, bool debugMode);
  GeoPolyhedronGeneric resolveBoolean(const GeoPolyhedronGeneric* phA, const GeoPolyhedronGeneric* phB, int modeBoolean, bool debugMode);

  void collectFaceNumber(const GeoShape*, int& nFaces);

 private:

  // mutable unsigned int m_counterNumerical = 0;
  double m_tolerance = 1.e-5;
  // cache input polyhedra
  const GeoPolyhedronGeneric* m_phA{nullptr};
  const GeoPolyhedronGeneric* m_phB{nullptr};  

  // method to find (distance to) the intersection of a (polyhedron) edge with (another) polyhedron face plane
  // arguments : edge: (1) starting point, (2) endpoint-starting point, plane: (3) point , (4) normal
  double edgeFaceIntersection( GeoTrf::Vector3D eInit, GeoTrf::Vector3D edge, GeoTrf::Vector3D facePoint, GeoTrf::Vector3D faceNormal ) const;  

  // auxiliary method : input edges
  std::vector<NPH_edge> collectEdges(std::vector<NPH_plane>& planes, std::vector<NPH_vertex>& vlog, bool debugMode = false ) const;

  // auxiliary method : intersections
  void collectLines( std::vector<NPH_plane>& planes, std::vector<GeoTrf::Vector3D>& vtx, std::vector<NPH_vertex>& vlog,
				     std::vector<NPH_edge>& edges, bool debugMode ) const;

  std::vector<NPL_edge> evaluateVertexLogic(int plane_index, std::vector<NPH_vertex> vlog, std::vector<NPH_edge> edges,
					    std::vector<NPH_plane>& planes, std::vector<GeoTrf::Vector3D> vtx) const;

  bool insideFace(unsigned int plane_index, GeoTrf::Vector3D point, double precision = 1.e-5, bool debugMode = false) const {
    unsigned int nPA = m_phA->getNFaces();
    return (   plane_index < m_phA->getNFaces() ? m_phA->insideFace(plane_index,point, precision, debugMode) :
	       m_phB->insideFace(plane_index-nPA, point, precision, debugMode)  ) ;
  }

  void intersectFace(unsigned int in_edge, std::vector<NPH_edge>& edges, unsigned int plane_index, std::vector<NPH_plane>& planes, std::vector<GeoTrf::Vector3D>& vtx, std::vector<NPH_vertex>& vlog) const;
  
  bool intersectLines(GeoTrf::Vector3D A, GeoTrf::Vector3D a, GeoTrf::Vector3D B, GeoTrf::Vector3D b, double& da, double& db, double& d) const;

  bool inPlane(NPH_vertex vlog, std::vector<NPH_edge> edges, int ip) const;

  NPL_vertex filterLog(NPH_vertex vlog, std::vector<NPH_edge> edges, int ip, std::vector<NPH_plane> planes) const;
  void filterLogPlane(NPL_vertex& plv, NPH_vertex& vlog, std::vector<NPH_edge> edges, int ip, std::vector<NPH_plane> planes) const;
  
  bool matchPlanes( int , int, std::vector<NPH_plane> planes ) const;

  std::vector<int> findHitPlane( NPH_vertex vlog, int ph_index, std::vector<NPH_edge> edges, std::vector<NPH_plane> planes) const;

  std::pair<int,int> hitDoubleEdge(int vx, int plane_index, int ph_index, std::vector<NPH_vertex> vlog, std::vector<NPH_edge> edges) const;

  void setDoubleEdge(int ie,int ia, std::vector<NPH_edge>& edges) const;
  
  NPL_edge symmetrize( NPL_edge line ) const;
  
  void saveHitOnEdge(NPH_vtx nvtx, std::vector<NPH_edge>& edges) const;
  
  NPL_edge createLine(int iv1, int iv2, int adj, NPH_plane plane) const;
  
  NPL_ordered mergeLoops( NPL_ordered loop1, NPL_ordered loop2, std::vector< GeoTrf::Vector3D > vtx);
  
  GeoTrf::Vector3D areaVec( std::vector<size_t> loop, std::vector< GeoTrf::Vector3D > vtx) const;

  int nextOri(int input, std::vector<NPL_edge> plane_edges, std::vector<NPL_edge> edge_loop_origin, std::vector<int> availableEdges, bool previous = false) const; 

  // for debugging purpose only
  std::vector<std::pair<int,int> > missing_edges( const GeoPolyhedronGeneric gpg) const;

};

#endif
