/*
  Copyright (C) 2002-2017 CERN for the benefit of the ATLAS collaboration
*/

#include "GeoModelKernel/GeoPolyhedronGeneric.h"
#include "GeoModelKernel/GeoShapeAction.h"
#include "GeoModelKernel/GeoBox.h"
#include "GeoModelKernel/GeoTrd.h"
#include "GeoModelKernel/GeoTube.h"
#include "GeoModelKernel/GeoTubs.h"
#include "GeoModelKernel/GeoPcon.h"
#include "GeoModelKernel/GeoSimplePolygonBrep.h"
#include "GeoModelKernel/GeoShapeShift.h"
#include <cmath>
#include <stdexcept>
#include <iostream>
#include <map>


const std::string GeoPolyhedronGeneric::s_classType = "PolyhedronGeneric";
const ShapeType GeoPolyhedronGeneric::s_classTypeID = 0x25;

GeoPolyhedronGeneric::GeoPolyhedronGeneric(const GeoShape* shape, const GeoTrf::Transform3D& transform, int polyhedrize) {

  std::vector<GeoTrf::Vector3D> vtx;
  m_transform = transform;

  if (shape->type() == "Box") {
    const GeoBox* box = dynamic_cast<const GeoBox*> (shape); 
    double hx = box->getXHalfLength();
    double hy = box->getYHalfLength();
    double hz = box->getZHalfLength();
      
    vtx.push_back(GeoTrf::Vector3D(-hx,-hy, hz));
    vtx.push_back(GeoTrf::Vector3D(-hx,+hy, hz));
    vtx.push_back(GeoTrf::Vector3D(+hx,+hy, hz));
    vtx.push_back(GeoTrf::Vector3D(+hx,-hy, hz));
    vtx.push_back(GeoTrf::Vector3D(+hx,-hy,-hz));
    vtx.push_back(GeoTrf::Vector3D(+hx,+hy,-hz));
    vtx.push_back(GeoTrf::Vector3D(-hx,+hy,-hz));
    vtx.push_back(GeoTrf::Vector3D(-hx,-hy,-hz));

    m_origin = 1;
  }

  if (shape->type() == "Trd") {
    const GeoTrd* trd = dynamic_cast<const GeoTrd*> (shape); 
    double hx1 = trd->getXHalfLength1();
    double hx2 = trd->getXHalfLength2();
    double hy1 = trd->getYHalfLength1();
    double hy2 = trd->getYHalfLength2();
    double hz = trd->getZHalfLength();
      
    vtx.push_back(GeoTrf::Vector3D(-hx2,-hy2, hz));
    vtx.push_back(GeoTrf::Vector3D(-hx2,+hy2, hz));
    vtx.push_back(GeoTrf::Vector3D(+hx2,+hy2, hz));
    vtx.push_back(GeoTrf::Vector3D(+hx2,-hy2, hz));
    vtx.push_back(GeoTrf::Vector3D(+hx1,-hy1,-hz));
    vtx.push_back(GeoTrf::Vector3D(+hx1,+hy1,-hz));
    vtx.push_back(GeoTrf::Vector3D(-hx1,+hy1,-hz));
    vtx.push_back(GeoTrf::Vector3D(-hx1,-hy1,-hz));

    m_origin = 2;
  }

  if (shape->type() == "Box" || shape->type() == "Trd") {

    addVertices(vtx);
    std::vector<size_t> f0 = {0,1,2,3};
    addFace(f0);
    std::vector<size_t> f1 = {4,5,6,7};
    addFace(f1);
    std::vector<size_t> f2 = {0,7,6,1};
    addFace(f2);
    std::vector<size_t> f3 = {3,4,7,0};
    addFace(f3);
    std::vector<size_t> f4 = {2,5,4,3};
    addFace(f4);
    std::vector<size_t> f5 = {1,6,5,2};
    addFace(f5);

    for (unsigned int i=0; i<getNFaces(); i++) 
      m_Normals.push_back(
	((m_Vertices[m_Faces[i][0]]-m_Vertices[m_Faces[i][1]]).cross(m_Vertices[m_Faces[i][2]]-m_Vertices[m_Faces[i][1]])).normalized() );

    m_status = isValid() ? 0 : 2;
    //if (!isValid() ) std::cout << " GeoPolyhedronGeneric constructor from Box/Trd not valid" << std::endl;

  } else if (shape->type() == "SimplePolygonBrep") {

    const GeoSimplePolygonBrep* spb = dynamic_cast<const GeoSimplePolygonBrep*> (shape); 
    double hz = spb->getDZ();
    unsigned int nvtx = spb->getNVertices();
      
    for (unsigned int i=0; i < nvtx; i++) vtx.push_back(GeoTrf::Vector3D(spb->getXVertex(i),spb->getYVertex(i),hz));  
    for (unsigned int i=0; i < nvtx; i++) vtx.push_back(GeoTrf::Vector3D(spb->getXVertex(i),spb->getYVertex(i),-hz));

    // consistent coding requires points to be ordered clockwise
    double clockwise = 0.;
    for (unsigned int iv = 0; iv < nvtx; iv++) {
      unsigned int iv1 = iv < nvtx-1 ? iv + 1 : iv + 1 - nvtx;
      unsigned int iv2 = iv < nvtx-2 ? iv + 2 : iv + 2 - nvtx;
      double pn = ( (vtx[iv]-vtx[iv1]).cross(vtx[iv2]-vtx[iv1]).normalized() ).z();
      if ( std::abs(pn) > 1.e-3 )  clockwise += pn;
    }

    addVertices(vtx);
        
    std::vector<size_t> f0; std::vector<size_t> f1;
    if (clockwise>0) {
      for (unsigned int i=0; i < nvtx; i++) f0.push_back(i);
      for (unsigned int i=2*nvtx-1; i >= nvtx; i--) f1.push_back(i);
    } else {
      for (int i=nvtx-1; i >= 0; i--) f0.push_back(i);
      for (int i=nvtx; i < 2*nvtx; i++) f1.push_back(i);
    }
    addFace(f0); addNormal(GeoTrf::Vector3D(0.,0.,1.));
    addFace(f1); addNormal(GeoTrf::Vector3D(0.,0.,-1.));
    
    for (unsigned int i=0; i<nvtx; i++) {
      std::vector<size_t> f; f.clear();
      if (clockwise>0) {
	f.push_back( i == nvtx-1 ? 0 : i+1); f.push_back(i); f.push_back(nvtx+i); f.push_back( i == nvtx-1 ? nvtx : nvtx+i+1);
      } else {
	f.push_back( i == nvtx-1 ? nvtx : nvtx+i+1); f.push_back(nvtx+i); f.push_back(i); f.push_back( i == nvtx-1 ? 0 : i+1); 
      }
      addFace(f); addNormal( (m_Vertices[f[0]]-m_Vertices[f[1]]).cross(m_Vertices[f[2]]-m_Vertices[f[1]]).normalized() );  
    }

    fillSpans();

    // double-check normals and consistent ordering of faces
    for (unsigned int ip=2; ip<getNFaces(); ip++) {
      if (m_Normals[ip].norm()!=1) {
	for (unsigned int iv=1; iv<m_Faces[ip].size()-2; iv++) {
	  // recalculate normal using other vertices
	  GeoTrf::Vector3D tn = ((m_Vertices[m_Faces[ip][iv]]-m_Vertices[m_Faces[ip][iv+1]]).cross(m_Vertices[m_Faces[ip][iv+2]]-m_Vertices[m_Faces[ip][iv+1]])).normalized();
	  if (tn.norm()>0) { m_Normals[ip] = tn; break; }
	}
      }
    }

    m_status = isValid() ? 0 : 2;
    //if (!isValid() ) std::cout << " GeoPolyhedronGeneric constructor from SimplePolygonBrep not valid" << std::endl;
    
    m_origin = 3;
  }
  else if ( (shape->type() == "Tube" || shape->type() == "Tubs") ) {   

    m_origin = 4;

    if (polyhedrize == 0) {
      m_status = 1; return;      
    }

    const GeoTube* tube = (shape->type() == "Tube") ? dynamic_cast<const GeoTube*> (shape) : nullptr;
    const GeoTubs* tubs = (shape->type() == "Tubs") ? dynamic_cast<const GeoTubs*> (shape) : nullptr;
    
    double hz = tube ? tube->getZHalfLength() : tubs->getZHalfLength();
    double rmin = tube ? tube->getRMin() : tubs->getRMin();
    double rmax = tube ? tube->getRMax() : tubs->getRMax();
    double sPhi = tubs ? tubs->getSPhi() : 0.;
    double dPhi = tubs ? tubs->getDPhi() : 2*acos(-1.);
    bool sector = dPhi < 2*acos(-1.);
    
    int nsplit = dPhi>acos(-1.) ? polyhedrize : int(0.5*polyhedrize);
    if ( nsplit < 2 ) nsplit = 2; 
    
    // build outer shell
    double ff = 0.5*dPhi/nsplit;
    double rx = rmax * sqrt( ff / sin(ff) / cos(ff) ); 
    for (unsigned int i=0; i < nsplit; i++) vtx.push_back(GeoTrf::Vector3D(rx*cos(sPhi+dPhi-i*2*ff),rx*sin(sPhi+dPhi-i*2*ff),hz));  

    std::vector<size_t> f0; 
    for (unsigned int i=0; i < nsplit; i++) f0.push_back(i);
    
    // sector needs the closing vertex,too
    if ( sector ) {
      f0.push_back(vtx.size()); vtx.push_back(GeoTrf::Vector3D(rx*cos(sPhi),rx*sin(sPhi),hz)); 
      if (!(rmin>0) ) {  // need the central point
	f0.push_back(vtx.size()); vtx.push_back(GeoTrf::Vector3D(0.,0.,hz)); 
      }      
    } else if (rmin>0)  f0.push_back(0); // needed for combined loop

    int nv = vtx.size();   // outer ring vertex count
    if (rmin>0) {      // inner ring mirrored
      double rn = rmin * sqrt( ff / sin(ff) / cos(ff) ); 
      for (int i = nv-1; i >= 0; i--) {
	vtx.push_back(GeoTrf::Vector3D(rn/rx*vtx[i].x(),rn/rx*vtx[i].y(),hz));
      }
      f0.push_back(2*nv-1);
      for (unsigned int i = nv; i< 2*nv; i++) f0.push_back(i);
    }
    addFace(f0); addNormal(GeoTrf::Vector3D(0.,0.,1.));

    int mv = vtx.size();
    
    // mirror the face
    int ntop = f0.size();
    for (unsigned int i=0; i<mv; i++) {
      vtx.push_back(GeoTrf::Vector3D(vtx[i].x(),vtx[i].y(),-hz));
    }
    std::vector<size_t> f1;
    for (int i=ntop-1; i>=0; i--) f1.push_back(f0[i]+mv); 
    addFace(f1); addNormal(GeoTrf::Vector3D(0.,0.,-1.));

    addVertices(vtx);
    
    for (unsigned int i=0; i<ntop; i++) {
      if (rmin>0 && !sector && (i==nsplit || i==ntop-1) ) continue;  // dummy lateral faces
      std::vector<size_t> f; f.clear();
      f.push_back( i == ntop-1 ? f0[0] : f0[i+1]); f.push_back(f0[i]); f.push_back(f0[i]+mv); f.push_back( i == ntop-1 ? f0[0]+mv : f0[i+1]+mv);
      addFace(f); addNormal( (m_Vertices[f[0]]-m_Vertices[f[1]]).cross(m_Vertices[f[2]]-m_Vertices[f[1]]).normalized() );
    }

    fillSpans();

    m_status = isValid() ? 0 : 2;
    //if (!isValid() ) std::cout << " GeoPolyhedronGeneric constructor from Tube/Tubs not valid" << std::endl;

  }
  else if ( shape->type() == "Pcon") {   
   
    m_origin = 5;

    if ( polyhedrize == 0 ) {
      m_status = 1; return;
    }
    
    const GeoPcon* pcon = dynamic_cast<const GeoPcon*> (shape);

    double sPhi = pcon->getSPhi();
    double dPhi = pcon->getDPhi();
    bool sector = dPhi < 2*acos(-1.);
    unsigned int nplanes = pcon->getNPlanes();
    double pord = ( pcon->getZPlane(0) < pcon->getZPlane(nplanes-1) ) ? 1. : -1.;
   
    int nsplit = dPhi>acos(-1.) ? polyhedrize : int(0.5*polyhedrize);
    if ( nsplit < 2 ) nsplit = 2; 

    double ff = 0.5*dPhi/nsplit;

    std::vector<size_t> f; int nv = 0; int ntop = 0; int mv = 0;

    for ( unsigned int iz = 0; iz < nplanes; iz++) {
      
      double z = pcon->getZPlane(iz);
      double rmin = pcon->getRMinPlane(iz);
      double rmax = pcon->getRMaxPlane(iz);
    
      // build outer shell
      double rx = rmax * sqrt( ff / sin(ff) / cos(ff) ); 
      for (unsigned int i=0; i < nsplit; i++) vtx.push_back(GeoTrf::Vector3D(rx*cos(sPhi+dPhi-i*2*ff),rx*sin(sPhi+dPhi-i*2*ff),z));  

      if (sector) {     
	vtx.push_back(GeoTrf::Vector3D(rx*cos(sPhi),rx*sin(sPhi),z));  // sector needs the closing vertex,too
	if (!(rmin>0) )  vtx.push_back(GeoTrf::Vector3D(0.,0.,z));   //  the central point needed
      }
      
      if (iz==0 || iz == nplanes-1 ) {
	f.clear();
	for (unsigned int i=0; i < nsplit; i++) f.push_back(vtx.size()-nsplit+i);    
        if ( sector ) {
	  f.push_back(vtx.size()); 
	  if (!(rmin>0) ) f.push_back(vtx.size());
	} else if (rmin>0)  f.push_back(vtx.size()-nsplit); // needed for combined loop
	if (iz==0) nv = vtx.size();   // outer ring vertex count
      }      

      // inner ring mirrored
      if (rmin>0) {   
	double rn = rmin * sqrt( ff / sin(ff) / cos(ff) ); 
	for (int i = vtx.size()-1; i >= vtx.size()-nv; i--) {
	  vtx.push_back(GeoTrf::Vector3D(rn/rx*vtx[i].x(),rn/rx*vtx[i].y(),z));
	}

	if (iz==0 || iz == nplanes-1) {
	  f.push_back(vtx.size()-1);
	  for (unsigned int i = vtx.size()-nv; i< vtx.size(); i++) f.push_back(i);
	}
      }

      if (iz==0) {
	ntop = f.size();
	mv = vtx.size();
      }

      if (iz==0 || iz == nplanes-1) {
	if ( (pord>0 && iz==nplanes-1) || (pord<0 && iz==0) ) {
	  addFace(f);  addNormal(GeoTrf::Vector3D(0.,0.,+1.));
	} else {
	  std::vector<size_t> fi;
	  for (int i=ntop-1; i>=0; i--) fi.push_back(f[i]);
	  addFace(fi);  addNormal(GeoTrf::Vector3D(0.,0.,-1.));
	}
      }
	
      if ( iz>0 ) {

	int nvx = vtx.size();
	for (unsigned int i=0; i<ntop; i++) {
	  if (rmin>0 && !sector && (i==nsplit || i==ntop-1) ) continue;  // dummy lateral faces
	  f.clear();
	  f.push_back( i == ntop-1 ? nvx-ntop : nvx-ntop+i+1); f.push_back(nvx-ntop+i);
		       f.push_back(nvx-2*ntop+i); f.push_back( i == ntop-1 ? nvx-2*ntop : nvx-2*ntop+i+1);
	  addFace(f); addNormal( (vtx[f[0]]-vtx[f[1]]).cross(vtx[f[2]]-vtx[f[1]]).normalized() );
	}
      }
    }
    
    addVertices(vtx);

    fillSpans();

    m_status = isValid() ? 0 : 2;
    //if (!isValid() ) std::cout << " GeoPolyhedronGeneric constructor from Pcon not valid" << std::endl;

  } else m_status = 1;
  
  if (m_face_span.size()<m_Faces.size()) fillSpans();

}

GeoPolyhedronGeneric::GeoPolyhedronGeneric(const GeoPolyhedronGeneric& gpg) {

  m_Vertices = gpg.m_Vertices;
  m_Faces = gpg.m_Faces;
  m_Normals = gpg.m_Normals;
  m_origin = gpg.m_origin;   // only used for debugging of initial conversion from GeoShape
  m_status = gpg.m_status;

  if (isValid()) fillSpans();

}

double GeoPolyhedronGeneric::volume () const {
  if (!isValid())
    throw std::runtime_error ("Volume requested for incomplete polyhedron");
  //
  GeoTrf::Vector3D refVtx = m_Vertices[0];  // reference point
  double vol = 0.; 
  std::vector<GeoTrf::Vector3D> vtx;

  for (unsigned int i=0; i< getNFaces(); i++) {

    // retrieve vertices
    std::vector<size_t> faceDef = getFace(i);
    vtx.clear();
    for (auto vi : faceDef ) vtx.push_back(m_Vertices[vi]);
    // normal
    GeoTrf::Vector3D normal = faceNormal(i);

    // area
    int n = vtx.size();
    GeoTrf::Vector3D area = vtx[0].cross(vtx[n - 1]);
    for (int k = 1; k < n; ++k) area += vtx[k].cross(vtx[k-1]);
    vol += 0.5 * area.norm() * (vtx[1]-refVtx).dot(normal) / 3.;
  }
  
  return std::abs(vol);
}

void GeoPolyhedronGeneric::extent (double& xmin, double& ymin, double& zmin,
                                    double& xmax, double& ymax, double& zmax) const
{
  if (!isValid())
    throw std::runtime_error ("Extent requested for incomplete polyhedron");
  xmin = xmax = m_Vertices[0].x();
  ymin = ymax = m_Vertices[0].y();
  zmin = zmax = m_Vertices[0].z();
  for (size_t k = 1; k < m_Vertices.size(); ++k)
  {
    double x = m_Vertices[k].x();
    double y = m_Vertices[k].y();
    double z = m_Vertices[k].z();
    xmin = std::min(xmin, x);
    xmax = std::max(xmax, x);
    ymin = std::min(ymin, y);
    ymax = std::max(ymax, y);
    zmin = std::min(zmin, z);
    zmax = std::max(zmax, z);
  }
}

bool GeoPolyhedronGeneric::contains (double x, double y, double z) const
{
  return this->contains( GeoTrf::Vector3D(x,y,z) );
}

bool GeoPolyhedronGeneric::contains ( GeoTrf::Vector3D point, double precision,  bool debugMode ) const
{
  // TODO : make the method independent from the direction of normals ( so that it can be used for their validation )
  if (debugMode) std::cout <<"test inside for point:" << point.x()<<"," <<point.y()<<"," << point.z()
			   <<":distance from ref:"<< ( point - getRefPoint() ).norm() <<" compare with span:" << getRefSpan() <<  std::endl;

  // compare with object extent
  if ( ( point - getRefPoint() ).norm() > getRefSpan() ) return false;
  
  // algorithm : count intersection with faces along probe line : point+(0,0,1)*t
  // odd/even number of intersections : inside/outside
  
  bool ineg = false; bool ipos = false;
  GeoTrf::Vector3D probeDir(0.,0.,1.);
  std::vector<GeoTrf::Vector3D> intersections;     // keep trace of intersections to intervene when an edge is hit

  for (auto vi : m_Vertices) if ( (vi - point).norm() < precision) return true;

  if (debugMode) for (auto vi : m_Vertices) std::cout <<"input vertices:" << vi.x() <<"," << vi.y() <<"," << vi.z() << std::endl;  
  
  double h,pn,dist;
  for (size_t i=0; i< getNFaces() ; i++) {
    std::vector<size_t> face= getFace(i);
    if (face.size() < 3) continue;
    // check if point lays within the face
    GeoTrf::Vector3D n = this->faceNormal(i);

    h = ( m_Vertices[m_Faces[i][0]] - point).dot(n);    // nearest distance point - plane
    
    if (  std::abs(h) < precision ) {
      if ( insideFace(i, point, precision) ) {
	if (debugMode) std::cout << "point at face:"<< i << std::endl;
	return true;
      }
    }
       
    pn = probeDir.dot(n);                       // dot product normal*probe direction
    //if (debugMode) std::cout <<" face normal:" << i <<":" << n.x() <<"," << n.y()<<"," << n.z() << std::endl;
    if ( std::abs(pn) > 1.e-9 ) {
      dist = h / pn;
      GeoTrf::Vector3D ix = point + dist*probeDir ;

      if ( insideFace(i, GeoTrf::Vector3D( ix ), precision, debugMode ) ) {
        //if (debugMode)
	//  std::cout << "plane intersection:" << (point+dist*probeDir).x() <<"," <<(point+dist*probeDir).y() <<"," <<(point+dist*probeDir).z()
	//	    <<": at distance:" << dist<<":valid:" << validIntersection << std::endl;
	if ( std::abs(dist) < precision ) {
	  if (debugMode) std::cout << "point at face:"<< i << std::endl;
	  return true;
	}
	bool edgeHit = false;
	for ( auto px : intersections ) if ((ix - px).norm() < precision ) edgeHit = true;
	if (!edgeHit) {
	  if (dist > precision) ipos ^= true;
	  else if (dist < -precision) ineg ^= true;
	  if (debugMode) std::cout << "valid intersection with face:" << i << ":at distance:" << dist <<":pos:" << ipos <<":neg:" << ineg << std::endl;
	  intersections.push_back(ix);
	}
      }
    }
  }

  return ipos && ineg;   // TODO check if not compatible
}

void GeoPolyhedronGeneric::addVertices(std::vector<GeoTrf::Vector3D> vertices)
{
  m_Vertices = vertices;

  bool isIdentity = true;
  if (m_transform.translation().norm() > std::numeric_limits<float>::epsilon() ) isIdentity = false;
  if (isIdentity) {
    const GeoTrf::RotationMatrix3D& rot = m_transform.linear();
    for (int i=0; i<3; i++) if ( std::abs(GeoTrf::Vector3D::Unit(i).dot(rot*GeoTrf::Vector3D::Unit(i)) - 1 ) > std::numeric_limits<float>::epsilon() ) isIdentity = false;
  }

  if (!isIdentity) for ( int i = 0; i < m_Vertices.size(); i++) m_Vertices[i] = m_transform * m_Vertices[i]; 
}

void GeoPolyhedronGeneric::addFace(std::vector<size_t> faceVtx)
{
  m_Faces.push_back(faceVtx);
}

void GeoPolyhedronGeneric::addNormal(GeoTrf::Vector3D n)
{
  m_Normals.push_back(n);
}


GeoTrf::Vector3D GeoPolyhedronGeneric::faceNormal(unsigned int i) const
{

  if (!m_Normals.size() ) {   // recalculate normals
    for (unsigned int i=0; i<getNFaces(); i++) {
      m_Normals.push_back(
        ( (m_Vertices[m_Faces[i][0]]-m_Vertices[m_Faces[i][1]]).cross(m_Vertices[m_Faces[i][2]]-m_Vertices[m_Faces[i][1]])).normalized() ) ;
      // check orientation 
      for (unsigned int iv=0; iv<m_Faces[i].size()-2; iv++) {
	GeoTrf::Vector3D ref = 0.5*(m_Vertices[m_Faces[i][iv]] + 0.5*(m_Vertices[m_Faces[i][iv+1]]+m_Vertices[m_Faces[i][iv+2]]));
	if (insideFace(i,ref)) {
	  bool cn = contains( ref+1.e-3*m_Normals.back());
	  if ( cn != contains( ref-1.e-3*m_Normals.back()) ) {   // valid test point
	    if (cn) m_Normals.back() = -1.* m_Normals.back(); 
	    break;
	  }
	}
      }
    }    
  }
  
  return m_Normals[i];
}

bool GeoPolyhedronGeneric::insideFace(unsigned int i, GeoTrf::Vector3D point, double precision, bool debugMode) const
{  
  std::vector<GeoTrf::Vector3D> vtx;
  std::vector<size_t> face= getFace(i);
  for (auto vi : face ) vtx.push_back(m_Vertices[vi]);

  if ( std::abs(( point - vtx[0]).dot(faceNormal(i) ) ) > precision ) {
    if (debugMode) std::cout << "insideFace called for point not on plane" << std::endl;
    return false;
  }

  GeoTrf::Vector3D probe = getRefPoint(i) - point ;   

  // check span
  if ( probe.norm() > getRefSpan() ) {
    if (debugMode) std::cout << "insideFace called for point outside span" << std::endl;
    return false;
  }

  // vertex
  for (auto vx : vtx) if ( (vx - point).norm()< precision ) {
      if (debugMode) std::cout << "insideFace vertex hit" << std::endl;
      return true;
    }

  // save valid intersections in order to handle case when probe hits a vertex ( double intersection )
  std::vector<GeoTrf::Vector3D> valid_intersections;
   
  if ( probe.norm() < precision) return true;  // on edge

  bool in = false;
  double d,dist,t;

  if (debugMode) {
    std::cout << "insideFace:" << i <<":test point:" << point.x() <<"," << point.y() <<"," << point.z() << std::endl;
    std::cout << "insideFace:probe:" << probe.x() <<"," << probe.y() <<"," << probe.z() <<":norm:" << probe.norm()<< std::endl;
    for ( auto iv : vtx ) std::cout << "face vertices:" << iv.x() <<"," << iv.y() <<"," << iv.z() << std::endl;
  }
  
  for (unsigned int iv = 0; iv < vtx.size(); iv++) {
    
    GeoTrf::Vector3D edge = (iv<vtx.size()-1) ? vtx[iv+1] - vtx[iv] : vtx[0] - vtx[iv];
    double ee = edge.norm();
    if (debugMode) std::cout << "insideFace:edge:" << edge.x() <<"," << edge.y() <<"," << edge.z() <<":norm:" << ee<< std::endl;

    // dist and t in units (probe, edge)
    bool lineIntersect = intersectLines( point, probe, vtx[iv], edge, dist, t, d);
    if (!lineIntersect && std::abs(d) > precision ) continue;

    if (!lineIntersect) { // point on edge line
      double td = ( point - vtx[iv] ).dot(edge)/ee/ee;
      if (td*ee > -precision && (td-1.)*ee < precision) return true;  // does not count as intersection otherwise
    }
    if (!lineIntersect) continue;
    //
    if (debugMode) std::cout << "insideFace:intersecting line:" <<iv <<":d:"<<  d << ":dist:" << dist <<":t:" << t << std::endl;
    if ( std::abs(d)< 1.e-3 && t*ee >- precision && (t-1.)*ee < precision ) {     // TODO calculate error on d to remove hardcoded tolerance
      //
      if (fabs(dist*probe.norm()) < precision ) return true;
      // check if not found already
      GeoTrf::Vector3D xn = vtx[iv]+t*edge;
      bool known = false; for ( auto x : valid_intersections ) if ( (xn -x).norm() < precision ) known = true;
      if (!known) {
	in ^= dist>0;
	valid_intersections.push_back(xn);
	if (debugMode) std::cout <<"insideFace:collecting valid intersection:" << xn.x() <<"," << xn.y() <<"," << xn.z()<< ":insideFace current:" << in  << std::endl;
      }
      if ( (xn-point).norm() < precision ) return true;  // point on edge 
    }

  }

  if (debugMode) std::cout << "insideFace returns:" << in << std::endl;

  return in;
  
}

void GeoPolyhedronGeneric::exec(GeoShapeAction *action) const
{
  action->handleShape(this);
}


bool GeoPolyhedronGeneric::intersectLines(GeoTrf::Vector3D A, GeoTrf::Vector3D a, GeoTrf::Vector3D B, GeoTrf::Vector3D b, double& da, double& db, double& dx) const
{
  // lines parametrization : A + a*da , B + b*db ; a,b define "units" of distances da, db
  // return value : closest distance between lines in native [mm] units
  
  // scalars
  double ab = a.dot(b); double a2 = a.dot(a); double b2 = b.dot(b);
  double ABa = (A-B).dot(a);
  double ABb = (A-B).dot(b);


  // special cases
  if ( std::abs(ab)/sqrt(a2*b2)<1.e-9 ) {     // to ensure sufficient precision
    da = -ABa / a2;
    db = ABb / b2;
    dx = (A + da*a -B - db*b).norm();
    return true;
  }

  if ( a2*b2 == ab*ab ) {     // parallel lines
    da = 0; db = 0;
    dx = ((A-B).cross(b)).norm()/b.norm();
    return false;       
  }

  da = (ABb*ab -b2*ABa) / (a2*b2 - ab*ab);
  db = (ABa + a2*da) / ab;
  dx = (A + da*a -B - db*b).norm();
  return true;
  
}

// this is obsolete
std::vector< std::pair< double, std::pair<size_t,size_t> > > GeoPolyhedronGeneric::edgeFaceIntersect( int i, GeoTrf::Vector3D eInit, GeoTrf::Vector3D edge ) const
{
  std::vector< std::pair< double, std::pair<size_t,size_t> > > output;

  std::vector<GeoTrf::Vector3D> vtx;
  std::vector<size_t> face= getFace(i);
  for (auto vi : face ) vtx.push_back(m_Vertices[vi]);

  double d,dist,t; 
  for (unsigned int iv = 0; iv < vtx.size(); iv++) {

    std::pair<size_t,size_t> fvx(face[iv],  (iv<face.size()-1) ? face[iv+1] : face[0]);
      
    GeoTrf::Vector3D fi = (iv<vtx.size()-1) ? vtx[iv+1] - vtx[iv] : vtx[0] - vtx[iv];

    // dist and t in units (probe, edge)
    bool lineIntersect = intersectLines( vtx[iv], fi, eInit, edge, dist, t, d);

    if (lineIntersect) {
      if ( std::abs(d)<1.e-6 && dist>=0 && dist <= 1 && t>=0. && t<=1.) {
	std::pair< double, std::pair<size_t,size_t> > ix(t, fvx);
	output.push_back(ix);
      }
    }
  }
  return output;
}


bool GeoPolyhedronGeneric::isValid( bool debugMode ) const
{
  // if (getNFaces()<4) return false;
  if ( m_status > 0 ) return false;

  // loop over edges, check that each appears with its mirrored version ( attached faces )
  for ( unsigned int ic=0; ic < getNFaces(); ic++) {
    std::vector<size_t> vtx = getFace(ic);
    if (vtx.size()<2) continue;      // should not happen though
    for (unsigned int iv=0; iv<vtx.size(); iv++) {
      std::pair<size_t,size_t> edge(vtx[iv], iv<vtx.size()-1 ? vtx[iv+1] : vtx[0]);  // probe
      bool mirrored = false;  // search mirrored edge
      for ( unsigned int icm=0; icm < getNFaces(); icm++) {
	std::vector<size_t> vtm = getFace(icm);
	if (vtm.size()<2) continue;       // should not happen though
	for (unsigned int ivm=0; ivm<vtm.size(); ivm++) {
	  std::pair<size_t,size_t> edgm(vtm[ivm], ivm<vtm.size()-1 ? vtm[ivm+1] : vtm[0]);  // test
	  if ( edgm.first == edge.second && edgm.second == edge.first) mirrored = true;
	  if (mirrored) break;
	} // end loop over other edges
	if (mirrored) break;
      } // end loop over other faces
      if (!mirrored) {
	if (debugMode) std::cout << "problem in polyhedron, edge not mirrored:face:edge:" << ic<<":" <<iv << std::endl;
	return false;
      }
    }    
  }

  return true;   
}

void GeoPolyhedronGeneric::fillSpans() {

  // calculate reference points and span for volumes
  m_cog = GeoTrf::Vector3D(0.,0.,0.); m_span = 0.;
  if (m_Vertices.size()) {
    for ( auto vx : m_Vertices ) m_cog = m_cog + vx; m_cog = 1./m_Vertices.size()*m_cog;
    for ( auto vx : m_Vertices ) m_span = std::max( m_span, (vx-m_cog).norm() );
  }
  // and faces
  m_face_cog.clear(); m_face_span.clear();
  for ( auto fi : m_Faces ) { GeoTrf::Vector3D fc(0.,0.,0.); double fs=0.; 
    if (fi.size()) {
      for ( auto vi : fi ) fc = fc + m_Vertices[vi]; fc = 1./fi.size() * fc; 
      for ( auto vi : fi) fs = std::max( fs, (m_Vertices[vi] - fc).norm() ); 
    }
    m_face_cog.push_back(fc); m_face_span.push_back(fs);
  }
}

GeoTrf::Vector3D GeoPolyhedronGeneric::getRefPoint(int face) const {

  // first check that info available
  if ( m_face_cog.size() < m_Faces.size() ) {
    std::cout << "WARNING: GeoPolyhedronGeneric object not constructed: run ::construct() first"<< std::endl;
    return GeoTrf::Vector3D(1.e5,1.e5,1.e5);  // dummy return value
  }

  if (face<0) return m_cog;
  else if ( face < m_face_cog.size() ) return m_face_cog[face];
  else return GeoTrf::Vector3D(1.e5,1.e5,1.e5);    // return dummy

}

double GeoPolyhedronGeneric::getRefSpan(int face) const {

  // first check that info available
  if ( m_face_span.size() < m_Faces.size() ) {
    std::cout << "WARNING: GeoPolyhedronGeneric object not constructed: run ::construct() first"<< std::endl;
    return 1.e5;  // dummy return value
  }

  if (face<0) return m_span;
  else if ( face < m_face_span.size() ) return m_face_span[face];
  else return 1.e5;    // return dummy

}

int GeoPolyhedronGeneric::getOrigin() const { return m_origin; }

void GeoPolyhedronGeneric::setOrigin( int ori ) { m_origin = ori; }

void GeoPolyhedronGeneric::cleanAfterBooleanOperation() {

  // remove dummy faces ( less than 3 vertices )
  std::vector<std::vector<size_t> > nfaces;
  std::vector< GeoTrf::Vector3D  > normals;
  for (unsigned ip = 0; ip < m_Faces.size(); ip++) {
    if (m_Faces[ip].size()>2) {
      nfaces.push_back(m_Faces[ip]);
      normals.push_back(m_Normals[ip]);
    }
  }
      
  // removing unused vertices  
  std::vector<GeoTrf::Vector3D> nvtx;
  std::map<size_t,size_t> m;
  for (unsigned ip = 0; ip < nfaces.size(); ip++) { 
    for (unsigned iv = 0; iv < nfaces[ip].size(); iv++ ) {
      auto const& pair = m.try_emplace(nfaces[ip][iv], nvtx.size());
      if (pair.second) nvtx.push_back(m_Vertices[nfaces[ip][iv]]);
      nfaces[ip][iv] = (*pair.first).second;
    }
  }

  m_Vertices = nvtx;
  m_Faces = nfaces;
  m_Normals = normals;

  fillSpans();
}

bool GeoPolyhedronGeneric::construct(bool debugMode) {

  fillSpans();
  m_origin = 0;
  m_status = isValid(debugMode) ? 0 : 2;
  return isValid(debugMode); 
  
}
